from tdm.abstract.datamodel import AbstractDirective
from tdm.datamodel.directives import CreateAccountDirective, CreateConceptDirective, CreatePlatformDirective
from tdm.v0.json_schema.directive import AbstractDirectiveModel, CreateAccountDirectiveModel, CreateConceptDirectiveModel, \
    CreatePlatformDirectiveModel


def convert_directive(directive: AbstractDirectiveModel) -> AbstractDirective:
    if isinstance(directive, CreateConceptDirectiveModel):
        return CreateConceptDirective(
            name=directive.name,
            concept_type=directive.concept_type,
            filters=directive.filters,
            notes=directive.notes,
            markers=directive.markers,
            access_level=directive.access_level,
            id=directive.id
        )
    elif isinstance(directive, CreateAccountDirectiveModel):
        return CreateAccountDirective(
            key=directive.key,
            platform_key=directive.platform_key,
            name=directive.name,
            url=directive.url
        )
    elif isinstance(directive, CreatePlatformDirectiveModel):
        return CreatePlatformDirective(
            key=directive.key,
            name=directive.name,
            platform_type=directive.platform_type,
            url=directive.url
        )



__all__ = [
    'AbstractMarkup', 'FrozenMarkup', 'AbstractMarkupModel', 'unfreeze'
]

from ._helper import unfreeze
from .abstract import AbstractMarkup
from .frozen import FrozenMarkup
from .model import AbstractMarkupModel



__all__ = [
    'CompositeValueType', 'ConceptType',
    'PropertyType', 'RelationPropertyType',
    'RelationType', 'SlotType', 'AtomValueType'
]

from ._composite import CompositeValueType
from ._concept import ConceptType
from ._property import PropertyType, RelationPropertyType
from ._relation import RelationType
from ._slot import SlotType
from ._value import AtomValueType



from collections import defaultdict
from typing import ClassVar, Dict, Optional, Tuple, Type

from pydantic import BaseModel

from tdm.abstract.datamodel import AbstractDomain, AbstractFact, AbstractValue, TalismanDocument
from tdm.datamodel.document import TalismanDocumentFactory
from tdm.legacy import DocumentMetadataModel
from .content import TreeDocumentContentModel
from .convert import build_structure, convert_concept_fact, convert_directive, convert_property_fact, convert_relation_fact, \
    convert_value_fact, get_metadata_facts
from .directive import CreateConceptDirectiveModel, DirectiveModel
from .fact import ConceptFactModel, FactModel, FactType, PropertyFactModel, RelationFactModel, ValueFactModel


class TalismanDocumentModel(BaseModel):
    _DEFAULT_DOMAIN: ClassVar[AbstractDomain] = None
    _domain: AbstractDomain = None

    id: str
    content: TreeDocumentContentModel
    metadata: Optional[DocumentMetadataModel]
    facts: Optional[Tuple[FactModel, ...]]  # pydantic fails to serialize FrozenSet[FactModel]
    directives: Optional[Tuple[DirectiveModel, ...]]

    @classmethod
    def set_default_domain(cls, domain: Optional[AbstractDomain]) -> None:
        cls._DEFAULT_DOMAIN = domain

    def set_domain(self, domain: Optional[AbstractDomain]) -> None:
        self._domain = domain

    def to_doc(self, titles: Dict[str, Tuple[str, str]], value_types: Dict[str, Type[AbstractValue]]) -> TalismanDocument:
        result = TalismanDocumentFactory(self._domain or self._DEFAULT_DOMAIN).create_document(id_=self.id)

        nodes, structure, node_links = build_structure(self.content)
        result = result.with_nodes(nodes).with_structure(structure).with_node_links(node_links, update=True).with_main_root(self.content.id)

        # convert document metadata to facts
        metadata = self.metadata.to_metadata() if self.metadata else {}
        result = result.with_metadata(metadata).with_facts(get_metadata_facts(self.id, metadata), update=True)

        # now convert facts
        old_facts = defaultdict(set)
        for fact in (self.facts or []):
            old_facts[fact.fact_type].add(fact)

        facts = []
        facts_mapping: dict[str, AbstractFact] = {}
        id2directive = {directive.id: directive for directive in self.directives or {}}
        # convert all CreateAccount and CreateAccount directives into new types
        directives = [convert_directive(d) for d in filter(lambda d: not isinstance(d, CreateConceptDirectiveModel), self.directives or {})]

        # start with concept facts
        # old concept fact consists of <status, type_id, value, mention> and should be converted to chain
        # ConceptFact(type_id) <- Property(Name) -> AtomValueFact(value) <-  MentionFact -> TextNodeMention(mention)
        for old_concept_fact in old_facts.get(FactType.CONCEPT, []):
            old_concept_fact: ConceptFactModel
            concept_fact, *other = convert_concept_fact(old_concept_fact, result, titles, value_types)
            facts_mapping[old_concept_fact.id] = concept_fact
            facts.extend((concept_fact, *other))
            # if concept fact exists for directive old CreateConceptDirectiveModel convert in CreateConceptDirective
            if old_concept_fact.value in id2directive:
                directives.append(convert_directive(id2directive[old_concept_fact.value]))

        # then convert ValueFacts
        # old value fact consists of <status, type_id, value, mention> and should be converted to chain
        # AtomValueFact(value) <- MentionFact -> TextNodeMention(mention)
        for old_value_fact in old_facts.get(FactType.VALUE, []):
            old_value_fact: ValueFactModel
            value_fact, *other = convert_value_fact(old_value_fact, result, value_types)
            facts_mapping[old_value_fact.id] = value_fact
            facts.extend((value_fact, *other))

        # now relations
        # old relations should be mapped to new ones with trivial modifications
        for old_relation_fact in old_facts.get(FactType.RELATION, []):
            old_relation_fact: RelationFactModel
            relation_fact = convert_relation_fact(old_relation_fact, facts_mapping)
            facts_mapping[old_relation_fact.id] = relation_fact
            facts.append(relation_fact)

        # and finally properties and relation properties
        for old_property_fact in old_facts.get(FactType.PROPERTY, []):
            old_property_fact: PropertyFactModel
            property_fact = convert_property_fact(old_property_fact, facts_mapping)
            facts.append(property_fact)

        result = result.with_facts(facts, update=True)
        result = result.with_directives(directives, update=True)
        return result



from typing import Any, Dict, List, Mapping, Set

from .abstract import AbstractElementSerializer

_COLLECTIONS: Mapping[type, type] = {
    tuple: List,  # in other case Tuple[A, ...] is converted to Tuple[A]. Deserializes to tuple
    set: Set,
    list: List
}


class WrappedElementSerializer(AbstractElementSerializer):
    def __init__(self, real_type: type, serializer: AbstractElementSerializer):
        self._real_type = real_type
        self._typing_type = _COLLECTIONS[real_type]
        self._serializer = serializer

    def serialize(self, element):
        return self._real_type(self._serializer.serialize(e) for e in element)

    def deserialize(self, serialized, typed_id2element: Dict[type, Dict[str, Any]]):
        return self._real_type(self._serializer.deserialize(s, typed_id2element) for s in serialized)

    def field_type(self, element_type):
        return self._typing_type[self._serializer.field_type(element_type)]



from typing import Any, Type

from .abstract import AbstractElementModel, AbstractModelSerializer


class DataclassSerializer(AbstractModelSerializer[Any]):
    def serialize(self, element: Any) -> AbstractElementModel[Any]:
        return self.field_type(type(element)).serialize(element)

    def field_type(self, element_type: Type[Any]) -> Type[AbstractElementModel[Any]]:
        from tdm.abstract.json_schema.model import create_model_for_type
        return create_model_for_type(element_type)



from typing import Collection, Tuple, Type, TypeVar

from tdm.abstract.datamodel import AbstractDomainType, AbstractFact, AbstractMarkup, AbstractNode, AbstractNodeMention, AbstractValue, \
    BaseNodeMetadata
from tdm.helper import uniform_collection
from .abstract import AbstractElementSerializer
from .dataclass import DataclassSerializer
from .domain import DomainTypeSerializer
from .identifiable import IdSerializer
from .identity import IdentitySerializer
from .markup import MarkupSerializer
from .mention import NodeMentionSerializer
from .type_ import TypeSerializer
from .value import ValueSerializer
from .wrap import WrappedElementSerializer

_BASE_SERIALIZERS = {
    AbstractNode: IdSerializer(AbstractNode),
    AbstractFact: IdSerializer(AbstractFact),
    AbstractNodeMention: NodeMentionSerializer(),
    BaseNodeMetadata: DataclassSerializer(),
    AbstractValue: ValueSerializer(),
    AbstractMarkup: MarkupSerializer(),
    AbstractDomainType: DomainTypeSerializer()
}

_E = TypeVar('_E')
_S = TypeVar('_S')


def _prepare(serializer: AbstractElementSerializer[_E, _S], arg: Type[_E]) -> Tuple[Type[_S], _E]:
    return serializer.field_type(arg), serializer


def get_serializer(t: Type) -> Tuple[type, AbstractElementSerializer]:
    real_type, arg = uniform_collection(t)
    if real_type is type:
        return _prepare(TypeSerializer(), arg)
    if real_type is not None:
        if issubclass(real_type, Collection):
            wrapped_type, serializer = get_serializer(arg)
            return _prepare(WrappedElementSerializer(real_type, serializer), arg)
        raise TypeError
    possible_types = set(arg.mro()).intersection(_BASE_SERIALIZERS)
    if len(possible_types) == 1:
        return _prepare(_BASE_SERIALIZERS[possible_types.pop()], arg)
    if hasattr(arg, '__dataclass_fields__'):
        serializer = DataclassSerializer()
        return _prepare(DataclassSerializer(), arg)
    return _prepare(IdentitySerializer(arg), arg)



from itertools import chain
from typing import Any, Dict, Iterable

from tdm.abstract.datamodel import AbstractFact, FactStatus
from tdm.datamodel.facts import AtomValueFact, ConceptFact, PropertyFact, RelationFact
from tdm.datamodel.facts.concept import ConceptValue
from tdm.datamodel.values import DoubleValue, IntValue, StringValue


def get_metadata_facts(doc_id: str, metadata: Dict[str, Any]) -> Iterable[AbstractFact]:
    document_fact = ConceptFact(FactStatus.APPROVED, "document", ConceptValue(doc_id))
    result = [document_fact]
    metadata = metadata
    if not metadata:
        return result
    if 'parent_uuid' in metadata:
        result.append(
            RelationFact(
                FactStatus.APPROVED, "parent", document_fact,
                ConceptFact(FactStatus.APPROVED, "document", ConceptValue(metadata.pop('parent_uuid')))
            )
        )
    for key, value in metadata.items():
        value_facts = generate_value(value)
        result.extend(
            PropertyFact(FactStatus.APPROVED, key, document_fact, value_fact) for value_fact in value_facts
        )
    return result


def generate_value(val: Any) -> Iterable[AtomValueFact]:
    if isinstance(val, int):
        return AtomValueFact(FactStatus.APPROVED, 'int', IntValue(val)),
    if isinstance(val, str):
        return AtomValueFact(FactStatus.APPROVED, 'str', StringValue(val)),
    if isinstance(val, float):
        return AtomValueFact(FactStatus.APPROVED, 'float', DoubleValue(val)),
    if isinstance(val, Iterable):
        return tuple(chain.from_iterable(generate_value(v) for v in val))
    raise ValueError
